/* reflection.d -- stub of reflection support for the D programming language

  version 2006-12-03

  Copyright (C) 2006 Thomas Kühne <thomas@kuehne.cn>

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
*/

module cn.kuehne.reflection;

private import std.string;
private import std.stdio;
private import std.regexp;

version(linux){
	// DMD <-> GDC
	version = Unix;
}

version(GNU){
	version = internal_v1;
}else version(DigitalMars){
	private import std.compiler;
	static if(std.compiler.version_minor < 176){
		pragma(msg, "AAA");
		version = internal_v1;
	}else{
		pragma(msg, "BBB");
		version = internal_v2;
		version = issue_637;
	}
}else{
	static assert(0, "unknown compiler");
}

extern(C) void scan_binary();

extern(C){
	void new_symbol(char* name, void* value, size_t size, SymbolType type){
		Symbol s = new Symbol(name, value, size, type);
		symbols[s.mangled_name] = s;
	}
}

static this(){
	scan_binary();
}

Symbol[char[]] symbols;

enum SymbolType : ubyte{
	UNSPECIFIED	= 0,
	OBJECT		= 1,
	FUNCTION	= 2
}

enum CallingConvention{
	C,
	Pascal,
	Stdcall
}

struct Argument{
	size_t size;
	void* data;
}

struct Return{
	int	_int;
	long	_long;
	union{
		void* 	pointer;
		size_t	_size_t;
	}
	real	_real;
	creal	_creal;
}

class ReflectionException : Exception{
	this(char[] msg){
		super(msg);
	}
}

class SymbolException : ReflectionException{
	this(char[] msg){
		super(msg);
	}
}

class Symbol{
	char[] mangled_name;
	SymbolType type;
	void* start;
	size_t size;

	this(char* name, void* start, size_t size, SymbolType type){
		mangled_name = std.string.toString(name);
		this.start = start;
		this.size = size;
		this.type = type;
	}

	static void filter(inout Symbol[char[]] sym, char[] pattern, bool remove_match = false){
		filter(sym, new RegExp(pattern), remove_match);
	}

	static void filter(inout Symbol[char[]] sym, RegExp pattern, bool remove_match = false){
		if(remove_match){
			foreach(name; sym.keys.dup){
				if(pattern.match(name)){
					sym.remove(name);
				}
			}
		}else{
			foreach(name; sym.keys.dup){
				if(!pattern.match(name)){
					sym.remove(name);
				}
			}
		}
	}

	Return call(CallingConvention conv, Argument[] args){
		switch(conv){
			case CallingConvention.C:{
				return callC(start, args);
			}case CallingConvention.Pascal:{
				return callPascal(start, args);
			}default:{
				throw new SymbolException("unknown calling convention");
			}
		}
	}

	static Return callC(void* func, Argument[] arguments){
		return callCLike(func, arguments, false, true);
	}

	static Return callPascal(void* func, Argument[] arguments){
		return callCLike(func, arguments, true, false);
	}

	private static Return callCLike(void* func, Argument[] arguments,
			bool r_to_l, bool cleanStack)
	{
		static if(size_t.sizeof == 4){
			int _EAX_;
			int _EDX_;
			real _ST0_;
			real _ST1_;
			size_t* size;
			size_t* arg;
			size_t cleanup;
			if(arguments && arguments.length){
				size = new size_t[arguments.length + 1];
				arg = new size_t[arguments.length + 1];
	
				foreach(index, argument; r_to_l ? arguments : arguments.dup.reverse){
					size[index] = argument.size;
					arg[index] = cast(size_t) argument.data;
				}
			}else{
				size = new size_t[1];
				arg = new size_t[1];
			}
			size[arguments.length] = 0; // end marker
			arg[arguments.length] = 0; // end marker


			/*
			 * don't change:
			 *	EBX, ESI, EDI, EBP
			 *	DS, ES, SS
			 */
			asm{
			next_argument:
				// get argument size
				mov EAX, size;
				add size, 4;
				mov EDX, [EAX];
				// last argument ?
				cmp EDX, 0;
				je do_call;
				// get data
				mov EAX, arg;
				add arg, 4;
				mov ECX, [EAX];
				// check argument size 
				cmp EDX, 4;
				jbe still_more;
				// argument is bigger than 32bits
				mov EAX, EDX;
				and EAX, 0xFF_FF_FF_F8;
				sub EAX, 4;
				add ECX, EAX;
			still_more:
				cmp EDX, 1;
				jl next_argument;
				mov EAX, [ECX];
				push EAX;
				sub ECX, 4;
				add cleanup, 4;
				sub EDX, 4;
				jmp still_more;
			do_call:
				call func;
				// adjust stack
				cmp cleanStack, 0;
				je no_clean;
				add ESP, cleanup;
			no_clean:
				// store results
				mov _EAX_, EAX;
				mov _EDX_, EDX;
				fstp _ST0_;
				fstp _ST1_;
			}

			Return result;
			result._size_t = _EAX_;
			result._int = _EAX_;
			result._real = _ST0_;
			result._creal = _ST1_ + (1i * _ST0_);
			result._long = cast(uint) _EDX_;
			result._long <<= 32;
			result._long |= cast(uint) _EAX_;
			return result;
		}else{
			static assert(0, "unimplemented pointer size");
		}
	}


	char[] toString(){
		return format("%s [start:%8s size:%4s type:%s]", mangled_name, start, size, .toString(type));
	}
}

Symbol findSymbol(void* addr){
	foreach(s; symbols){
		if((s.start <= addr) && (addr < s.start + s.size)){
			return s;
		}
	}
	return null;
}

char[] addressToString(void* addr){
	Symbol s = findSymbol(addr);
	if(s){
		if(s.start == addr){
			return s.mangled_name;
		}else{
			return format("%s+0x%X", s.mangled_name, addr - cast(size_t)s.start);
		}
	}
	return format("0x%08X", addr);
}

char[] toString(SymbolType t){
	const char[][] v = ["UNSPECIFIED", "OBJECT", "FUNCTION"];
	return v[t];
}

CallingConvention toCallingConvention(char[] name){
	switch(name){
		case "C", "cdecl", "_cdecl", "__cdecl":
			return CallingConvention.C;
		case "Pascal", "__pascal":
			return CallingConvention.Pascal;
		default:
			throw new Exception("unsupported calling convention '"~ name ~"'");
	}
}

private extern(C) Object _d_newclass(ClassInfo);

/**
 * generate a new instance of a given class
 */
Object new_Instance(char[] name){
	Symbol* s = null;
	ClassInfo ci = null;
	void* object = null;

	char[] mangled;
	char[] mangled_info;
	char[] mangled_new;
	char[] mangled_this;

	// mangle class name
	foreach(element; std.string.split(name, ".")){
		mangled = format("%s%s%s", mangled, element.length, element);
	}

	// mangle internal data
	version(internal_v2){
		mangled_info = "_D" ~ mangled ~ "7__ClassZ";
		mangled_new = "_D" ~ mangled ~ "3newFkZPv";
		mangled_this = "_D" ~ mangled ~ "5_ctorFZC" ~ mangled;
	}else version(internal_v1){
		mangled_info = "_Class_" ~ mangled;	
		mangled_new = "_D" ~ mangled ~ "3newFkZPv";
		mangled_this = "_D" ~ mangled ~ "5_ctorFZC" ~ mangled;
	}else{
		static assert(0, "unknown mangling of internal data");
	}

	version(issue_637){
		pragma(msg, "deploying #637 work-around"); 
		mangled_info = mangled_info[1 .. $];
	}

	// find classinfo
	s = mangled_info in symbols;
	if (s is null){
		throw new ReflectionException("no classinfo found for " ~ name);
	}
	ci = cast(ClassInfo) s.start;

	// allocate
	s = mangled_new in symbols;
	if(s is null){
		// standard allocator
		object = _d_newclass(ci);
	}else{
		// custom allocator
		object = (cast(void* function(uint))s.start)(ci.sizeof);
		(cast(byte*)object)[0 .. ci.init.length] = ci.init[];
	}

	// call constructor
	s = mangled_this in symbols;
	if(s is null){
		throw new ReflectionException("no constructor found for " ~ name);
	}

	(cast(void function(void*))s.start)(object);

	return cast(Object)object;
}
